package http

import (
	"net/http"
	"time"

	"amuz.es/gogs/infra/changer/http/middleware/logging"
	"amuz.es/gogs/infra/changer/http/middleware/recovery"
	"amuz.es/gogs/infra/changer/http/middleware/session"
	"amuz.es/gogs/infra/changer/http/route"
	"amuz.es/gogs/infra/changer/http/template"
	"amuz.es/gogs/infra/changer/util"
	"github.com/gin-gonic/gin"
	errs "github.com/pkg/errors"
	graceful "gopkg.in/tylerb/graceful.v1"
)

var (
	logger = util.NewLogger("http")
	server *graceful.Server
)

func InitHttp(
	bindAddress string,
	prefix string,
	accessLogConfig *util.LogConfig,
	ldapConfig *util.LdapConfig,
	sessionConfig *util.SessionConfig,
	profile bool,
	errorSignal chan error, closeSignal chan struct{}) {
	gin.SetMode(gin.ReleaseMode)
	engine := gin.New()

	session.InitRole()

	initMiddleware(engine, accessLogConfig,ldapConfig, sessionConfig)

	initRoutes(engine, prefix, profile)
	server = &graceful.Server{
		Timeout: 10 * time.Second,
		Server: &http.Server{
			Addr:    bindAddress,
			Handler: engine,
		},
		TCPKeepAlive:     3 * time.Second,
		NoSignalHandling: true,
		LogFunc:          logger.Debugf,
	}
	go bind(server, errorSignal, closeSignal)
	logger.Info(bindAddress + " bound")
}

func CloseHttp() {
	logger.Info("HTTP closing..")
	server.Stop(10 * time.Second)
	logger.Info("HTTP closed")
}
func bind(server *graceful.Server, errorSignal chan error, closeSignal chan struct{}) {
	if err := server.ListenAndServe(); err != nil {
		errorSignal <- errs.Wrap(err, "failed to listen or serve http")
	} else {
		close(closeSignal)
	}
}
func initRoutes(engine *gin.Engine, prefix string, profile bool) {
	suburl := prefix
	path := engine.Group(suburl)

	path.StaticFS("/static", route.Static(""))


	if profile {
		logger.Info("enable Pprof handler ")
		path.GET("/debug/pprof/", IndexHandler())
		path.GET("/debug/pprof/heap", HeapHandler())
		path.GET("/debug/pprof/goroutine", GoroutineHandler())
		path.GET("/debug/pprof/block", BlockHandler())
		path.GET("/debug/pprof/threadcreate", ThreadCreateHandler())
		path.GET("/debug/pprof/cmdline", CmdlineHandler())
		path.GET("/debug/pprof/profile", ProfileHandler())
		path.GET("/debug/pprof/symbol", SymbolHandler())
		path.POST("/debug/pprof/symbol", SymbolHandler())
		path.GET("/debug/pprof/trace", TraceHandler())
	}

	path.GET("/favicon.ico", func(c *gin.Context) {
		c.Redirect(http.StatusMovedPermanently, "/static/favicon/images/favicon.ico")
	})

	path.GET("/manifest.json", func(c *gin.Context) {
		c.Redirect(http.StatusMovedPermanently, "/static/favicon/manifest.json")
	})

	path.GET("/browserconfig.xml", func(c *gin.Context) {
		c.Redirect(http.StatusMovedPermanently, "/static/favicon/browserconfig.xml")
	})

	path.GET("/", route.Index)
	path.GET("/change_password", route.ChangePasswordView)
	path.POST("/change_password", route.ChangePassword)
	path.GET("/login", route.LoginView)
	path.POST("/login", route.Login)
	path.GET("/logout", route.Logout)
}

func initMiddleware(engine *gin.Engine, accessLogConfig *util.LogConfig,ldapConfig *util.LdapConfig, sessionConfig *util.SessionConfig) {
	engine.Use(logging.AccessLog(accessLogConfig))
	engine.Use(recovery.RecoveryJSON())
	engine.Use(session.New(ldapConfig, sessionConfig))

	engine.HTMLRender = template.DefaultRender()

	// not found handler
	engine.NoRoute(recovery.RecoveryHttpHTMLError(http.StatusNotFound))
	engine.NoMethod(recovery.RecoveryHttpHTMLError(http.StatusMethodNotAllowed))
}
