// middleware
package logging

import (
	"bytes"
	"io"
	"net"
	"net/http"
	"strconv"
	"time"

	"amuz.es/gogs/infra/changer/http/iface"
	"amuz.es/gogs/infra/changer/util"
	"github.com/gin-gonic/gin"
)

const (
	dateFormat = "02/Jan/2006:15:04:05 -0700"
)

var DurationAspect = iface.NewRequestDurationAggr()
var CounterAspect = iface.NewCounterMetric()

type logItem struct {
	host          string
	method        string
	uri           string
	proto         string
	referer       string
	userAgent     string
	requestedHost string
	requestedPath string
	status        int
	responseSize  int
	requestTime   time.Time
	duration      time.Duration
}

func AccessLog(config *util.LogConfig) gin.HandlerFunc {
	_, logWriter := util.NewLogWriter(config)
	DurationAspect.StartTimer(10 * time.Second)
	CounterAspect.StartTimer(1 * time.Minute)
	logChan := make(chan logItem)

	go sendLogging(logWriter, logChan)
	return func(c *gin.Context) {
		now := time.Now()
		req := c.Request

		defer func() {
			dur := time.Now().Sub(now)

			logItemData := logItem{
				status:       c.Writer.Status(),
				responseSize: c.Writer.Size(),
				requestTime:  now,
				duration:     dur,
			}

			if req != nil {
				logItemData.host = remoteHost(req)
				logItemData.method = req.Method
				logItemData.uri = req.RequestURI
				logItemData.proto = req.Proto
				logItemData.referer = req.Referer()
				logItemData.userAgent = req.UserAgent()
				logItemData.requestedHost = req.Host
				logItemData.requestedPath = req.URL.Path
			} else {
				logItemData.host = "-"
				logItemData.method = ""
				logItemData.uri = ""
				logItemData.proto = ""
				logItemData.referer = ""
				logItemData.userAgent = ""
				logItemData.requestedHost = ""
				logItemData.requestedPath = ""

			}

			logChan <- logItemData
		}()

		c.Next()
	}
}
func sendLogging(accesslog io.Writer, logChan chan logItem) {
	for {
		select {
		case logItemData := <-logChan:
			DurationAspect.Add(float64(logItemData.duration.Nanoseconds()))
			CounterAspect.Inc <- iface.Tuple{
				Path: logItemData.requestedPath,
				Code: logItemData.status,
			}
			// Logs an access event in Apache combined log format (with a minor customization with the duration).

			buf := bytes.NewBufferString("")
			buf.WriteString(logItemData.host)
			buf.WriteString(` - - [`)
			buf.WriteString(logItemData.requestTime.Format(dateFormat))
			buf.WriteString(`] "`)
			buf.WriteString(logItemData.method)
			buf.WriteByte(' ')
			buf.WriteString(logItemData.uri)
			buf.WriteByte(' ')
			buf.WriteString(logItemData.proto)
			buf.WriteString(`" `)
			buf.WriteString(strconv.Itoa(logItemData.status))
			buf.WriteByte(' ')
			buf.WriteString(strconv.Itoa(logItemData.responseSize))
			buf.WriteString(` "`)
			buf.WriteString(logItemData.referer)
			buf.WriteString(`" "`)
			buf.WriteString(logItemData.userAgent)
			buf.WriteString(`" `)
			buf.WriteString(strconv.FormatInt((logItemData.duration.Nanoseconds() / time.Millisecond.Nanoseconds()), 10))
			buf.WriteByte(' ')
			buf.WriteString(logItemData.requestedHost)
			buf.WriteByte('\n')
			accesslog.Write(buf.Bytes())
		}
	}
}

// strip port from addresses with hostname, ipv4 or ipv6
func stripPort(address string) string {
	if h, _, err := net.SplitHostPort(address); err == nil {
		return h
	}

	return address
}

// The remote address of the client. When the 'X-Forwarded-For'
// header is set, then it is used instead.
func remoteAddr(r *http.Request) string {
	ff := r.Header.Get("X-Forwarded-For")
	if ff != "" {
		return ff
	}

	return r.RemoteAddr
}

func remoteHost(r *http.Request) string {
	a := remoteAddr(r)
	h := stripPort(a)
	if h != "" {
		return h
	}

	return "-"
}
