package recovery

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"net/http"
	"net/http/httputil"
	"runtime"
	"strings"

	"amuz.es/gogs/infra/changer/http/iface"
	"amuz.es/gogs/infra/changer/util"
	"github.com/Sirupsen/logrus"
	"github.com/flosch/pongo2"
	"github.com/gin-gonic/gin"
)

var (
	dunno     = []byte("???")
	centerDot = []byte("·")
	dot       = []byte(".")
	slash     = []byte("/")
	logger    = util.NewLogger("recovery")
	color     = "danger"
	preText   = "루이보스 API 오류 :bomb:"
	title     = "API 요청을 처리하는 중에 의도하지 않은 오류가 발행하였습니다."
	footer    = "확인해 주세요! :hugging_face:"
)

func recoveryInternal(stack []byte, code int, req *http.Request, err *string) {
	httprequest, _ := httputil.DumpRequest(req, false)
	detailInfo := map[string]interface{}{
		"method":     req.Method,
		"statusCode": code,
		"path":       req.URL.Path,
		"message":    err,
		"request":    string(httprequest),
		"stack":      string(stack),
	}
	// message := slack.Payload{
	// 	Attachments: []slack.Attachment{
	// 		convertSlackMessage(code, err, &stack, req, &httprequest),
	// 	},
	// }
	// slack.Send(util.Config.Slack.Url, "", message)
	logger.WithFields(logrus.Fields(detailInfo)).Error("[Recovery] panic recovered:\n")
}

func recoveryError(code int, _ *gin.Context) *iface.WrappedResponse {
	return &iface.WrappedResponse{
		Message: http.StatusText(code),
	}
}

func RecoveryHttpJSONError(code int) gin.HandlerFunc {
	return func(c *gin.Context) {
		info := recoveryError(code, c)
		util.DumpJSON(c, code, info)
	}
}

func RecoveryHttpHTMLError(code int) gin.HandlerFunc {
	return func(c *gin.Context) {
		info := recoveryError(code, c)
		c.HTML(code, "views/error.html", pongo2.Context{
			"code":    code,
			"message": info.Message,
		})
	}
}

func RecoveryJSON() gin.HandlerFunc {
	return func(c *gin.Context) {
		defer func() {
			if err := recover(); err != nil {
				code := http.StatusInternalServerError
				stack := stack(5)
				systemErrMessage := err.(error).Error()
				go recoveryInternal(stack, code, c.Copy().Request, &systemErrMessage)
				util.DumpJSON(c, code, &iface.WrappedResponse{
					Message: systemErrMessage,
				})
				c.Abort()
			}
		}()
		c.Next()
	}
}

func RecoveryHTML() gin.HandlerFunc {
	return func(c *gin.Context) {
		defer func() {
			if err := recover(); err != nil {
				code := http.StatusInternalServerError
				systemErrMessage := err.(error).Error()
				c.HTML(code, "views/error.html", pongo2.Context{
					"code":    code,
					"message": systemErrMessage,
				})
				c.Abort()
			}
		}()
		c.Next()
	}
}

// stack returns a nicely formated stack frame, skipping skip frames
func stack(skip int) []byte {
	buf := new(bytes.Buffer) // the returned data
	// As we loop, we open files and read them. These variables record the currently
	// loaded file.
	var lines [][]byte
	var lastFile string
	for i := skip; ; i++ {
		// Skip the expected number of frames
		pc, file, line, ok := runtime.Caller(i)
		if !ok {
			break
		}
		if paths := strings.SplitN(file, "src/", 2); len(paths) == 1 {
			// Print this much at least.  If we can't find the source, it won't show.
			fmt.Fprintf(buf, "%s:%d (0x%x)\n", file, line, pc)
		} else if vendors := strings.SplitN(paths[1], "vendor/", 2); len(vendors) == 1 {
			// Print this much at least.  If we can't find the source, it won't show.
			fmt.Fprintf(buf, "%s:%d (0x%x)\n", paths[1], line, pc)
		} else {
			// Print this much at least.  If we can't find the source, it won't show.
			fmt.Fprintf(buf, "%s:%d (0x%x)\n", vendors[1], line, pc)
		}
		if file != lastFile {
			data, err := ioutil.ReadFile(file)
			if err != nil {
				continue
			}
			lines = bytes.Split(data, []byte{'\n'})
			lastFile = file
		}
		fmt.Fprintf(buf, "\t%s: %s\n", function(pc), source(lines, line))
	}
	return buf.Bytes()
}

// source returns a space-trimmed slice of the n'th line.
func source(lines [][]byte, n int) []byte {
	n-- // in stack trace, lines are 1-indexed but our array is 0-indexed
	if n < 0 || n >= len(lines) {
		return dunno
	}
	return bytes.TrimSpace(lines[n])

}

// function returns, if possible, the name of the function containing the PC.
func function(pc uintptr) []byte {
	fn := runtime.FuncForPC(pc)
	if fn == nil {
		return dunno
	}
	name := []byte(fn.Name())
	// The name includes the path name to the package, which is unnecessary
	// since the file name is already included.  Plus, it has center dots.
	// That is, we see
	//	runtime/debug.*T·ptrmethod
	// and want
	//	*T.ptrmethod
	// Also the package path might contains dot (e.g. code.google.com/...),
	// so first eliminate the path prefix
	if lastslash := bytes.LastIndex(name, slash); lastslash >= 0 {
		name = name[lastslash+1:]
	}
	if period := bytes.Index(name, dot); period >= 0 {
		name = name[period+1:]
	}
	name = bytes.Replace(name, centerDot, dot, -1)
	return name
}

// func convertSlackMessage(
// 	code int, err *string, stack *[]byte,
// 	req *http.Request, httpdumpreq *[]byte) (attachment slack.Attachment) {
// 	errmsg := "N/A"
// 	if err != nil {
// 		errmsg = *err
// 	}
// 	return slack.Attachment{
// 		Color:   &color,
// 		PreText: &preText,
// 		Title:   &title,
// 		Footer:  &footer,
// 		Text:    err,
// 		Fields: []*slack.Field{
// 			{Title: "status", Value: fmt.Sprintf("%s(%d)", http.StatusText(code), code), Short: true},
// 			{Title: "method", Value: req.Method, Short: true},
// 			{Title: "path", Value: req.URL.Path, Short: true},
// 			{Title: "phase", Value: util.Config.Phase, Short: true},
// 			{Title: "version", Value: util.Config.Version(), Short: true},
// 			{Title: "message", Value: errmsg, Short: true},
// 			{Title: "request", Value: string(*httpdumpreq), Short: false},
// 			{Title: "stack", Value: string(*stack), Short: false},
// 		},
// 	}
// }
