package template

import (
	"bytes"
	"fmt"
	"io"
	"path/filepath"
	"regexp"

	"github.com/flosch/pongo2"
)

const filePathNormalizerFormat = `^%s?(.*)%s?$`
const seperator = string(filepath.Separator)

type memoryTemplateLoader struct {
	loaderFunc     func(path string) ([]byte, error)
	pathNormalizer *regexp.Regexp
	fallbackLoader *pongo2.TemplateLoader
}

//NewMemoryTemplateLoader loads a bindata template loader
func NewMemoryTemplateLoader(loaderFunc func(path string) ([]byte, error)) pongo2.TemplateLoader {
	quotedSeperator := regexp.QuoteMeta(seperator)
	return &memoryTemplateLoader{
		loaderFunc:     loaderFunc,
		pathNormalizer: regexp.MustCompile(fmt.Sprintf(filePathNormalizerFormat, quotedSeperator, quotedSeperator)),
	}
}

//NewMemoryTemplateLoaderWithFallback loads a bindata template loader, but if the file is not found
//it will try to find it in the fallback loader with the same path
func NewMemoryTemplateLoaderWithFallback(loaderFunc func(path string) ([]byte, error), fallbackLoader pongo2.TemplateLoader) pongo2.TemplateLoader {
	return &memoryTemplateLoader{loaderFunc: loaderFunc, fallbackLoader: &fallbackLoader}
}

func (m memoryTemplateLoader) Abs(base, name string) string {
	if filepath.IsAbs(name) || base == "" {
		return m.pathNormalizer.FindStringSubmatch(filepath.Clean(name))[1]
	}

	if name == "" {
		return m.pathNormalizer.FindStringSubmatch(filepath.Clean(base))[1]
	}

	return m.pathNormalizer.FindStringSubmatch(filepath.Clean(filepath.Join(filepath.Dir(base), name)))[0]
}

func (m memoryTemplateLoader) Get(path string) (io.Reader, error) {
	data, err := m.loaderFunc(path)
	if err != nil {
		if m.fallbackLoader != nil {
			return (*m.fallbackLoader).Get(path)
		}

		return nil, err
	}

	return bytes.NewReader(data), nil
}
