package crypt

import (
	"bytes"
	"crypto/aes"
	"crypto/cipher"
	"crypto/rand"
	"encoding/base64"
	"fmt"
	"io"
	"strings"

	"github.com/pkg/errors"
)

func NewAesCryptor(key []byte) AesCryptor {
	return &aesCryptor{
		key,
		12,
	}
}

type AesCryptor interface {
	Encrypt(plaintext []byte) (data, nonce []byte, err error)
	Decrypt(encrypted, nonce []byte) (data []byte, err error)

	EncryptEncoded(plaintext []byte) (data string, err error)
	DecryptEncoded(encrypted string) (data []byte, err error)
}
type aesCryptor struct {
	key         []byte
	nonceLength int
}

func (c *aesCryptor) Encrypt(plaintext []byte) (data []byte, nonce []byte, err error) {
	defer func() {
		if r := recover(); r != nil {
			if rerr, ok := r.(error); !ok {
				err = fmt.Errorf("pkg: %v", r)
			} else {
				err = rerr
			}
		}
	}()

	var block cipher.Block
	block, err = aes.NewCipher(c.key)
	if err != nil {
		return
	}

	// Never use more than 2^32 random nonces with a given key because of the risk of a repeat.
	nonce = make([]byte, c.nonceLength)
	if _, err = io.ReadFull(rand.Reader, nonce); err != nil {
		return
	}
	var aesgcm cipher.AEAD
	aesgcm, err = cipher.NewGCM(block)
	if err != nil {
		return
	}

	data = aesgcm.Seal(nil, nonce, plaintext, nil)
	return
}

func (c *aesCryptor) Decrypt(encrypted []byte, nonce []byte) (data []byte, err error) {
	defer func() {
		if r := recover(); r != nil {
			if rerr, ok := r.(error); !ok {
				err = fmt.Errorf("pkg: %v", r)
			} else {
				err = rerr
			}
		}
	}()

	var block cipher.Block
	block, err = aes.NewCipher(c.key)
	if err != nil {
		return
	}

	var aesgcm cipher.AEAD
	aesgcm, err = cipher.NewGCM(block)
	if err != nil {
		return
	}

	data, err = aesgcm.Open(nil, nonce, encrypted, nil)
	return
}

func (c *aesCryptor) EncryptEncoded(plaintext []byte) (data string, err error) {
	defer func() {
		if r := recover(); r != nil {
			if rerr, ok := r.(error); !ok {
				err = fmt.Errorf("pkg: %v", r)
			} else {
				err = rerr
			}
		}
	}()
	var (
		encrypted []byte
		nonce     []byte
	)
	encrypted, nonce, err = c.Encrypt(plaintext)
	if err != nil {
		return
	}
	var buf bytes.Buffer
	encoder := base64.NewEncoder(base64.URLEncoding, &buf)
	if _, err = encoder.Write(encrypted); err != nil {
		return
	}

	if err = encoder.Close(); err != nil {
		return
	}
	buf.WriteRune('$')

	if _, err = encoder.Write(nonce); err != nil {
		return
	}

	if err = encoder.Close(); err != nil {
		return
	}
	data = buf.String()
	return
}
func (c *aesCryptor) DecryptEncoded(encryptedText string) (data []byte, err error) {

	defer func() {
		if r := recover(); r != nil {
			if rerr, ok := r.(error); !ok {
				err = fmt.Errorf("pkg: %v", r)
			} else {
				err = rerr
			}
		}
	}()
	splittedText := strings.SplitN(encryptedText, "$", 2)

	if len(splittedText) != 2 {
		err = errors.New("invalid data ")
		return
	}

	var (
		encrypted []byte
		nonce     []byte
	)

	if encrypted, err = 	base64.URLEncoding.DecodeString(splittedText[0]); err != nil {
		return
	}

	if nonce, err = base64.URLEncoding.DecodeString(splittedText[1]); err != nil{
		return
	}

	if len(nonce) != c.nonceLength {
		err = errors.Errorf("nonce length not %d ,actual %d", c.nonceLength, len(nonce))
		return
	}
	data, err = c.Decrypt(encrypted, nonce)
	return
}
