package iface

import (
	"bytes"
	"database/sql/driver"
	"encoding/hex"
	"errors"
	"fmt"

	"github.com/mailru/easyjson/jlexer"
	"github.com/mailru/easyjson/jwriter"
)

type Bytes16 [16]byte

func (b Bytes16) Marshal() ([]byte, error) {
	return b[:], nil
}

func (b Bytes16) MarshalTo(buf []byte) (n int, err error) {
	copy(buf, b[:])
	return len(b), nil
}
func (b *Bytes16) Unmarshal(buf []byte) error {
	if len(buf) != 16 {
		return fmt.Errorf("invalid bytes16 (got %d bytes)", len(buf))
	}
	copy(b[:], buf)
	return nil
}

func (b Bytes16) Compare(other Bytes16) int {
	return bytes.Compare(b[:], other[:])
}

func (b Bytes16) Equal(other Bytes16) bool {
	return b.Compare(other) == 0
}

// ParseBytes is like Parse, except it parses a byte slice instead of a string.
func ParseBytes16(buf []byte) (b Bytes16, err error) {
	if len(buf) != hex.EncodedLen(len(b)) {
		err = fmt.Errorf("invalid bytes16 length: %d", len(buf))
	}
	_, err = hex.Decode(b[:], buf)
	return
}

func (b Bytes16) MarshalEasyJSON(w *jwriter.Writer) {
	w.String(hex.EncodeToString(b[:]))
}
func (b *Bytes16) UnmarshalEasyJSON(w *jlexer.Lexer) {
	if parsed, err := ParseBytes16([]byte(w.String())); err == nil {
		*b = parsed
	}
}

func (b *Bytes16) Size() int {
	if b == nil {
		return 0
	}
	return 16
}

func (b *Bytes16) FromHexString(buf []byte) (error) {
	hexBuf := make([]byte, hex.DecodedLen(len(buf)))
	if _, err := hex.Decode(hexBuf, buf); err != nil {
		return err
	} else if err := b.Unmarshal(hexBuf); err != nil {
		return err
	}
	return nil
}

func (b Bytes16) ToHexString() (string) {
	return hex.EncodeToString(b[:])
}

// Scan implements the Scanner interface.
func (b *Bytes16) Scan(src interface{}) error {
	switch src.(type) {
	case string:
		b.FromHexString([]byte(src.(string)))
	default:
		return errors.New("Incompatible type for UUID")
	}
	return nil
}

// Value implements the driver Valuer interface.
func (b Bytes16) Value() (driver.Value, error) {
	return b.ToHexString(), nil
}
