package iface

import (
	"bytes"
	"encoding/hex"
	"fmt"

	"github.com/mailru/easyjson/jlexer"
	"github.com/mailru/easyjson/jwriter"
	"database/sql/driver"
	"errors"
)

type Bytes32 [32]byte

func (b Bytes32) Marshal() ([]byte, error) {
	return b[:], nil
}

func (b Bytes32) MarshalTo(buf []byte) (n int, err error) {
	copy(buf, b[:])
	return len(b), nil
}
func (b *Bytes32) Unmarshal(buf []byte) error {
	if len(buf) != 32 {
		return fmt.Errorf("invalid bytes32 (got %d bytes)", len(buf))
	}
	copy(b[:], buf)
	return nil
}

func (b Bytes32) Compare(other Bytes32) int {
	return bytes.Compare(b[:], other[:])
}

func (b Bytes32) Equal(other Bytes32) bool {
	return b.Compare(other) == 0
}

// ParseBytes is like Parse, except it parses a byte slice instead of a string.
func ParseBytes32(buf []byte) (b Bytes32, err error) {
	if len(buf) != hex.EncodedLen(len(b)) {
		err = fmt.Errorf("invalid bytes32 length: %d", len(buf))
	}
	_, err = hex.Decode(b[:], buf)
	return
}

func (b Bytes32) MarshalEasyJSON(w *jwriter.Writer) {
	w.String(hex.EncodeToString(b[:]))
}
func (b *Bytes32) UnmarshalEasyJSON(w *jlexer.Lexer) {
	if parsed, err := ParseBytes32([]byte(w.String())); err == nil {
		*b = parsed
	}
}

func (b *Bytes32) Size() int {
	if b == nil {
		return 0
	}
	return 32
}


func (b *Bytes32) FromHexString(buf []byte) (error) {
	hexBuf := make([]byte, hex.DecodedLen(len(buf)))
	if _, err := hex.Decode(hexBuf, buf); err != nil {
		return err
	} else if err := b.Unmarshal(hexBuf); err != nil {
		return err
	}
	return nil
}

func (b Bytes32) ToHexString() (string) {
	return hex.EncodeToString(b[:])
}

// Scan implements the Scanner interface.
func (b *Bytes32) Scan(src interface{}) error {
	switch src.(type) {
	case string:
		b.FromHexString([]byte(src.(string)))
	default:
		return errors.New("Incompatible type for UUID")
	}
	return nil
}

// Value implements the driver Valuer interface.
func (b Bytes32) Value() (driver.Value, error) {
	return b.ToHexString(), nil
}
