package iface

import (
	"bytes"
	"encoding/hex"
	"fmt"

	"github.com/mailru/easyjson/jlexer"
	"github.com/mailru/easyjson/jwriter"
	"database/sql/driver"
	"errors"
)

type Bytes64 [64]byte

func (b Bytes64) Marshal() ([]byte, error) {
	return b[:], nil
}

func (b Bytes64) MarshalTo(buf []byte) (n int, err error) {
	copy(buf, b[:])
	return len(b), nil
}
func (b *Bytes64) Unmarshal(buf []byte) error {
	if len(buf) != 64 {
		return fmt.Errorf("invalid bytes64 (got %d bytes)", len(buf))
	}
	copy(b[:], buf)
	return nil
}

func (b Bytes64) Compare(other Bytes64) int {
	return bytes.Compare(b[:], other[:])
}

func (b Bytes64) Equal(other Bytes64) bool {
	return b.Compare(other) == 0
}

// ParseBytes is like Parse, except it parses a byte slice instead of a string.
func ParseBytes64(buf []byte) (b Bytes64, err error) {
	if len(buf) != hex.EncodedLen(len(b)) {
		err = fmt.Errorf("invalid bytes64 length: %d", len(buf))
	}
	_, err = hex.Decode(b[:], buf)
	return
}

func (b Bytes64) MarshalEasyJSON(w *jwriter.Writer) {
	w.String(hex.EncodeToString(b[:]))
}
func (b *Bytes64) UnmarshalEasyJSON(w *jlexer.Lexer) {
	if parsed, err := ParseBytes64([]byte(w.String())); err == nil {
		*b = parsed
	}
}
func (b *Bytes64) Size() int {
	if b == nil {
		return 0
	}
	return 64
}


func (b *Bytes64) FromHexString(buf []byte) (error) {
	hexBuf := make([]byte, hex.DecodedLen(len(buf)))
	if _, err := hex.Decode(hexBuf, buf); err != nil {
		return err
	} else if err := b.Unmarshal(hexBuf); err != nil {
		return err
	}
	return nil
}

func (b Bytes64) ToHexString() (string) {
	return hex.EncodeToString(b[:])
}

// Scan implements the Scanner interface.
func (b *Bytes64) Scan(src interface{}) error {
	switch src.(type) {
	case string:
		b.FromHexString([]byte(src.(string)))
	default:
		return errors.New("Incompatible type for UUID")
	}
	return nil
}

// Value implements the driver Valuer interface.
func (b Bytes64) Value() (driver.Value, error) {
	return b.ToHexString(), nil
}
