package iface

import (
	"bytes"
	"database/sql/driver"
	"encoding/hex"
	"errors"
	"github.com/google/uuid"
	"github.com/mailru/easyjson/jlexer"
	"github.com/mailru/easyjson/jwriter"
)

type UUID struct {
	uuid.UUID
}

func (u UUID) Marshal() ([]byte, error) {
	return u.MarshalBinary()
}

func (u UUID) MarshalTo(buf []byte) (n int, err error) {
	if len(u.UUID) == 0 {
		return 0, nil
	}
	copy(buf, u.UUID[:])
	return len(u.UUID), nil
}
func (u *UUID) Unmarshal(buf []byte) error {
	return u.UnmarshalBinary(buf)
}

func (u UUID) Compare(other UUID) int {
	return bytes.Compare(u.UUID[:], other.UUID[:])
}

func (u UUID) Equal(other UUID) bool {
	return u.Compare(other) == 0
}

func (u UUID) MarshalEasyJSON(w *jwriter.Writer) {
	w.String(hex.EncodeToString(u.UUID[:]))
}
func (u *UUID) UnmarshalEasyJSON(w *jlexer.Lexer) {
	if decoded, err := hex.DecodeString(w.String()); err != nil {
		return
	} else if err := u.UnmarshalBinary(decoded); err != nil {
		return
	}
}

func (u *UUID) Size() int {
	if u == nil {
		return 0
	}
	if len((*u).UUID) == 0 {
		return 0
	}
	return 16
}

func NewUUIDV4() (u UUID, err error) {
	var ui uuid.UUID
	if ui, err = uuid.NewRandom(); err == nil {
		u.UUID = ui
	}
	return
}

func (u *UUID) UUIDFromHexString(buf []byte) (error) {
	hexBuf := make([]byte, hex.DecodedLen(len(buf)))
	if _, err := hex.Decode(hexBuf, buf); err != nil {
		return err
	} else if err := u.UnmarshalBinary(hexBuf); err != nil {
		return err
	}
	return nil
}

func (u UUID) ToHexString() (string) {
	return hex.EncodeToString(u.UUID[:])
}

// Scan implements the Scanner interface.
func (u *UUID) Scan(src interface{}) error {
	switch src.(type) {
	case string:
		u.UUIDFromHexString([]byte(src.(string)))
	default:
		return errors.New("Incompatible type for UUID")
	}
	return nil
}

// Value implements the driver Valuer interface.
func (u UUID) Value() (driver.Value, error) {
	return u.ToHexString(), nil
}

// Value implements the driver Valuer interface.
func (u UUID) Random() {
	u.UUID = uuid.New()
}
