package ldap

import (
	"fmt"
	"strings"

	"amuz.es/gogs/infra/changer/enums"
	"amuz.es/gogs/infra/changer/util"
	"crypto/tls"
	"github.com/pkg/errors"
	"gopkg.in/ldap.v2"
)

type SecurityProtocol int

// Note: new type must be added at the end of list to maintain compatibility.
const (
	SECURITY_PROTOCOL_UNENCRYPTED SecurityProtocol = iota
	SECURITY_PROTOCOL_LDAPS
	SECURITY_PROTOCOL_START_TLS
)

var logger = util.NewLogger("ldap")

// Basic LDAP authentication service
type ldapHandler struct {
	Host              string // LDAP host
	Port              int    // port number
	SecurityProtocol  SecurityProtocol
	SkipVerify        bool
	BindDN            string // DN to bind with
	BindPassword      string // Bind DN password
	BaseDN            string // Base search path for users
	UserDN            string // Template for the DN of the user for simple auth
	AttributeUsername string // Username attribute
	AttributeName     string // First name attribute
	AttributeSurname  string // Surname attribute
	AttributeMail     string // E-mail attribute
	AttributesInBind  bool   // fetch attributes in bind context (not user)
	Filter            string // Query filter to validate entry
	AdminFilter       string // Query filter to check if user is admin
}

type LdapAuthenticator interface {
	GetInfo(userId string) (map[string]interface{}, error)
	AuthenticateUser(name, passwd string, directBind bool) (error)
	SearchPermission(userId string) (enums.UserType, error)
	ChangePassword(name, passwd string, newPassword string, directBind bool) (error)
}

func NewLdapAuthenticator(
	host string,
	port int,
	securityProtocol int,
	skipVerify bool,
	bindDN string,
	bindPassword string,
	baseDN string,
	userDN string,
	attributeUsername string,
	attributeName string,
	attributeSurname string,
	attributeMail string,
	filter string,
	adminFilter string) LdapAuthenticator {
	return &ldapHandler{
		Host:              host,
		Port:              port,
		SecurityProtocol:  SecurityProtocol(securityProtocol),
		SkipVerify:        skipVerify,
		BindDN:            bindDN,
		BindPassword:      bindPassword,
		BaseDN:            baseDN,
		UserDN:            userDN,
		AttributeUsername: attributeUsername,
		AttributeName:     attributeName,
		AttributeSurname:  attributeSurname,
		AttributeMail:     attributeMail,
		Filter:            filter,
		AdminFilter:       adminFilter,
	}
}

func (ls *ldapHandler) sanitizedUserQuery(username string) (string, bool) {
	// See http://tools.ietf.org/search/rfc4515
	badCharacters := "\x00()*\\"
	if strings.ContainsAny(username, badCharacters) {
		logger.Debugf("Username contains invalid query characters: %s", username)
		return "", false
	}

	return fmt.Sprintf(ls.Filter, ldap.EscapeFilter(username)), true
}

func (ls *ldapHandler) sanitizedUserDN(username string) (string, bool) {
	// See http://tools.ietf.org/search/rfc4514: "special characters"
	badCharacters := "\x00()*\\,='\"#+;<>"
	if strings.ContainsAny(username, badCharacters) || strings.HasPrefix(username, " ") || strings.HasSuffix(username, " ") {
		logger.Debugf("Username contains invalid query characters: %s", username)
		return "", false
	}

	return fmt.Sprintf(ls.UserDN, ldap.EscapeFilter(username)), true
}

func (ls *ldapHandler) findUserDN(l *ldap.Conn, name string) (string, bool) {
	logger.Debugf("Search for LDAP user: %s", name)
	if ls.BindDN != "" && ls.BindPassword != "" {
		err := l.Bind(ls.BindDN, ls.BindPassword)
		if err != nil {
			logger.Debugf("Failed to bind as BindDN '%s': %v", ls.BindDN, err)
			return "", false
		}
		logger.Debugf("Bound as BindDN: %s", ls.BindDN)
	} else {
		logger.Debugf("Proceeding with anonymous LDAP search")
	}

	// A search for the user.
	userFilter, ok := ls.sanitizedUserQuery(name)
	if !ok {
		return "", false
	}

	logger.Debugf("Searching for DN using filter '%s' and base '%s'", userFilter, ls.BaseDN)
	search := ldap.NewSearchRequest(
		ls.BaseDN, ldap.ScopeWholeSubtree, ldap.NeverDerefAliases, 0, 0,
		false, userFilter, []string{}, nil)

	// Ensure we found a user
	sr, err := l.Search(search)
	if err != nil || len(sr.Entries) < 1 {
		logger.Debugf("Failed search using filter '%s': %v", userFilter, err)
		return "", false
	} else if len(sr.Entries) > 1 {
		logger.Debugf("Filter '%s' returned more than one user", userFilter)
		return "", false
	}

	userDN := sr.Entries[0].DN
	if userDN == "" {
		logger.Error(4, "LDAP search was successful, but found no DN!")
		return "", false
	}

	return userDN, true
}

func dial(ls *ldapHandler) (*ldap.Conn, error) {
	logger.Debugf("Dialing LDAP with security protocol '%v' without verifying: %v", ls.SecurityProtocol, ls.SkipVerify)

	tlsCfg := &tls.Config{
		ServerName:         ls.Host,
		InsecureSkipVerify: ls.SkipVerify,
	}
	if ls.SecurityProtocol == SECURITY_PROTOCOL_LDAPS {
		return ldap.DialTLS("tcp", fmt.Sprintf("%s:%d", ls.Host, ls.Port), tlsCfg)
	}

	conn, err := ldap.Dial("tcp", fmt.Sprintf("%s:%d", ls.Host, ls.Port))
	if err != nil {
		return nil, fmt.Errorf("Dial: %v", err)
	}

	if ls.SecurityProtocol == SECURITY_PROTOCOL_START_TLS {
		if err = conn.StartTLS(tlsCfg); err != nil {
			conn.Close()
			return nil, fmt.Errorf("StartTLS: %v", err)
		}
	}

	return conn, nil
}

func bindUser(l *ldap.Conn, userDN, passwd string) error {
	logger.Debugf("Binding with userDN: %s", userDN)
	err := l.Bind(userDN, passwd)
	if err != nil {
		logger.Debugf("LDAP authentication failed for '%s': %v", userDN, err)
		return err
	}
	logger.Debugf("Bound successfully with userDN: %s", userDN)
	return err
}

func changePassword(l *ldap.Conn, userDN, oldPasswd, newPasswd string) error {
	logger.Debugf("Change Password userDN: %s", userDN)

	passwordModifyRequest := ldap.NewPasswordModifyRequest(userDN, oldPasswd, newPasswd)

	if _, err := l.PasswordModify(passwordModifyRequest); err != nil {
		logger.Debugf("LDAP authentication failed for '%s': %v", userDN, err)
		return err
	}

	logger.Debugf("Change Password successfully with userDN: %s", userDN)
	return nil
}

// searchEntry : search an LDAP source if an entry (name, passwd) is valid and in the specific filter
func (ls *ldapHandler) SearchPermission(userId string) (enums.UserType, error) {
	// See https://tools.ietf.org/search/rfc4513#section-5.1.2
	l, err := dial(ls)
	if err != nil {
		logger.Errorf("LDAP connect failed for '%s': %v", ls.Host, err)
		return enums.UserTypeAnon, errors.Errorf("LDAP connect failed for '%s': %v", ls.Host, err)
	}
	defer l.Close()

	var userDN string
	logger.Debugf("LDAP will use BindDN")

	var found bool
	userDN, found = ls.findUserDN(l, userId)
	if !found {
		return enums.UserTypeAnon, errors.Errorf("cannot find user %s", userId)
	}

	userType := enums.UserTypeUser
	if len(ls.AdminFilter) > 0 {
		logger.Debugf("Checking admin with filter '%s' and base '%s'", ls.AdminFilter, userDN)
		search := ldap.NewSearchRequest(
			userDN, ldap.ScopeWholeSubtree, ldap.NeverDerefAliases, 0, 0, false, ls.AdminFilter,
			[]string{ls.AttributeName},
			nil)

		sr, err := l.Search(search)
		if err != nil {
			logger.Errorf("LDAP admin search failed: %v", err)
		} else if len(sr.Entries) < 1 {
			logger.Error("LDAP admin search failed: 0 entries")
		} else {
			userType = enums.UserTypeAdmin
		}
	}
	return userType, nil
}

// searchEntry : search an LDAP source if an entry (name, passwd) is valid and in the specific filter
func (ls *ldapHandler) GetInfo(userId string) (map[string]interface{}, error) {
	// See https://tools.ietf.org/search/rfc4513#section-5.1.2
	result := make(map[string]interface{})
	result["role"] = enums.UserTypeAnon

	l, err := dial(ls)
	if err != nil {
		logger.Errorf("LDAP connect failed for '%s': %v", ls.Host, err)
		return result, errors.Errorf("LDAP connect failed for '%s': %v", ls.Host, err)
	}
	defer l.Close()

	var userDN string
	logger.Debugf("LDAP will use BindDN")

	var found bool
	userDN, found = ls.findUserDN(l, userId)
	if !found {
		return result, errors.Errorf("cannot find user %s", userId)
	}

	userFilter, ok := ls.sanitizedUserQuery(userId)
	if !ok {
		return result, errors.Errorf("cannot evulate user(%s) query", userId)
	}

	logger.Debugf("Fetching attributes '%v', '%v', '%v', '%v' with filter '%s' and base '%s'", ls.AttributeUsername, ls.AttributeName, ls.AttributeSurname, ls.AttributeMail, userFilter, userDN)
	search := ldap.NewSearchRequest(
		userDN, ldap.ScopeWholeSubtree, ldap.NeverDerefAliases, 0, 0, false, userFilter,
		[]string{ls.AttributeUsername, ls.AttributeName, ls.AttributeSurname, ls.AttributeMail},
		nil)

	sr, err := l.Search(search)
	if err != nil {
		logger.Errorf("LDAP search failed: %v", err)
		return result, errors.Errorf("LDAP search failed: %v", err)
	} else if len(sr.Entries) < 1 {
		logger.Error("LDAP search failed: 0 entries")
		return result, errors.New("LDAP search failed: 0 entries")
	}

	result["username"] = sr.Entries[0].GetAttributeValue(ls.AttributeUsername)
	result["firstName"] = sr.Entries[0].GetAttributeValue(ls.AttributeName)
	result["surName"] = sr.Entries[0].GetAttributeValue(ls.AttributeSurname)
	result["mail"] = sr.Entries[0].GetAttributeValue(ls.AttributeMail)

	userType := enums.UserTypeUser
	if len(ls.AdminFilter) > 0 {
		logger.Debugf("Checking admin with filter '%s' and base '%s'", ls.AdminFilter, userDN)
		search := ldap.NewSearchRequest(
			userDN, ldap.ScopeWholeSubtree, ldap.NeverDerefAliases, 0, 0, false, ls.AdminFilter,
			[]string{ls.AttributeName},
			nil)

		sr, err := l.Search(search)
		if err != nil {
			logger.Errorf("LDAP admin search failed: %v", err)
		} else if len(sr.Entries) < 1 {
			logger.Error("LDAP admin search failed: 0 entries")
		} else {
			userType = enums.UserTypeAdmin
		}
	}
	result["role"] = userType

	return result, nil
}

// searchEntry : search an LDAP source if an entry (name, passwd) is valid and in the specific filter
func (ls *ldapHandler) AuthenticateUser(name, passwd string, directBind bool) (error) {
	// See https://tools.ietf.org/search/rfc4513#section-5.1.2
	if len(passwd) == 0 {
		logger.Debugf("authentication failed for '%s' with empty password", name)
		return errors.Errorf("authentication failed for '%s' with empty password", name)
	}
	l, err := dial(ls)
	if err != nil {
		logger.Errorf("LDAP connect failed for '%s': %v", ls.Host, err)
		return errors.Errorf("LDAP connect failed for '%s': %v", ls.Host, err)
	}
	defer l.Close()

	var userDN string
	if directBind {
		logger.Debugf("LDAP will bind directly via UserDN template: %s", ls.UserDN)

		var ok bool
		userDN, ok = ls.sanitizedUserDN(name)
		if !ok {
			return errors.Errorf("cannot find user %s", name)
		}
	} else {
		logger.Debugf("LDAP will use BindDN")

		var found bool
		userDN, found = ls.findUserDN(l, name)
		if !found {
			return errors.Errorf("cannot find user %s", name)
		}
	}

	if directBind || !ls.AttributesInBind {
		// binds user (checking password) before looking-up attributes in user context
		err = bindUser(l, userDN, passwd)
		if err != nil {
			return errors.New("password not match")
		}
	} else {
		// binds user (checking password) after looking-up attributes in BindDN context
		err = bindUser(l, userDN, passwd)
		if err != nil {
			return errors.New("password not match")
		}
	}
	return nil
}

// searchEntry : search an LDAP source if an entry (name, passwd) is valid and in the specific filter
func (ls *ldapHandler) ChangePassword(name, passwd string, newPassword string, directBind bool) (error) {
	// See https://tools.ietf.org/search/rfc4513#section-5.1.2
	if len(passwd) == 0 {
		logger.Debugf("authentication failed for '%s' with empty password", name)
		return errors.Errorf("authentication failed for '%s' with empty password", name)
	}
	l, err := dial(ls)
	if err != nil {
		logger.Errorf("LDAP connect failed for '%s': %v", ls.Host, err)
		return errors.Errorf("LDAP connect failed for '%s': %v", ls.Host, err)
	}
	defer l.Close()

	var userDN string
	if directBind {
		logger.Debugf("LDAP will bind directly via UserDN template: %s", ls.UserDN)

		var ok bool
		userDN, ok = ls.sanitizedUserDN(name)
		if !ok {
			return errors.Errorf("cannot find user %s", name)
		}
	} else {
		logger.Debugf("LDAP will use BindDN")

		var found bool
		userDN, found = ls.findUserDN(l, name)
		if !found {
			return errors.Errorf("cannot find user %s", name)
		}
	}

	if directBind || !ls.AttributesInBind {
		// binds user (checking password) before looking-up attributes in user context
		err = bindUser(l, userDN, passwd)
		if err != nil {
			return errors.New("password not match")
		}
	} else {
		// binds user (checking password) after looking-up attributes in BindDN context
		err = bindUser(l, userDN, passwd)
		if err != nil {
			return errors.New("password not match")
		}
	}

	err = changePassword(l, userDN, passwd, newPassword)

	if err != nil {
		return errors.New("cannot change password")
	}
	return err
}
