package util

import (
	"bytes"
	"crypto/rand"
	"fmt"
	"image"
	"math"
	"reflect"
	"strings"
	"time"
	"unsafe"
)

func Int64ToBytes(s int64) []byte {
	return []byte{
		byte(s >> 0),
		byte(s >> 8),
		byte(s >> 16),
		byte(s >> 24),
		byte(s >> 32),
		byte(s >> 40),
		byte(s >> 48),
		byte(s >> 56),
	}
}

func BytesToInt64(s []byte) int64 {
	return int64(0) |
		(int64(s[0]) << 0) |
		(int64(s[1]) << 8) |
		(int64(s[2]) << 16) |
		(int64(s[3]) << 24) |
		(int64(s[4]) << 32) |
		(int64(s[5]) << 40) |
		(int64(s[6]) << 48) |
		(int64(s[7]) << 56)
}

func BytesToString(b []byte) string {
	bh := (*reflect.SliceHeader)(unsafe.Pointer(&b))
	sh := reflect.StringHeader{
		Data: bh.Data,
		Len:  bh.Len}
	return *(*string)(unsafe.Pointer(&sh))
}

func StringToBytes(s string) []byte {
	sh := (*reflect.StringHeader)(unsafe.Pointer(&s))
	bh := reflect.SliceHeader{
		Data: sh.Data,
		Len:  sh.Len,
		Cap:  0}
	return *(*[]byte)(unsafe.Pointer(&bh))
}

// GetRandomString generate random string by specify chars.
func GetRandomString(n int, alphabets ...byte) string {
	const alphanum = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
	var bytes = make([]byte, 0, n)
	rand.Read(bytes)
	for _, b := range bytes {
		if len(alphabets) == 0 {
			bytes = append(bytes, alphanum[b%byte(len(alphanum))])
		} else {
			bytes = append(bytes, alphabets[b%byte(len(alphabets))])
		}
	}
	return string(bytes)
}

// Seconds-based time units
var (
	NanoSecond  uint64 = 1
	MicroSecond        = 1000 * NanoSecond
	MilliSecond        = 1000 * MicroSecond
	Second             = 1000 * MilliSecond
	Minute             = 60 * Second
	Hour               = 60 * Minute
	Day                = 24 * Hour
	Week               = 7 * Day
	Month              = 30 * Day
	Year               = 12 * Month
)

func computeTimeDiff(buff *bytes.Buffer, diff uint64) uint64 {
	switch {
	case diff < 1*NanoSecond:
		diff = 0
	case diff < 2*NanoSecond:
		diff = 0
		buff.WriteString("001ns")
	case diff < 1*MicroSecond:
		buff.WriteString(fmt.Sprintf("%03dns", diff))
		diff -= diff / NanoSecond * NanoSecond
	case diff < 2*MicroSecond:
		diff -= 1 * MilliSecond
		buff.WriteString("001us")
	case diff < 1*MilliSecond:
		buff.WriteString(fmt.Sprintf("%03dus", diff/MicroSecond))
		diff -= diff / MicroSecond * MicroSecond
	case diff < 2*MilliSecond:
		diff -= 1 * Second
		buff.WriteString("001ms")
	case diff < 1*Second:
		buff.WriteString(fmt.Sprintf("%03dms", diff/MilliSecond))
		diff -= diff / MilliSecond * MilliSecond
	case diff < 2*Second:
		diff -= 1 * Minute
		buff.WriteString("01s")
	case diff < 1*Minute:
		buff.WriteString(fmt.Sprintf("%02ds", diff/Second))
		diff -= diff / Second * Second
	case diff < 2*Minute:
		diff -= 1 * Minute
		buff.WriteString("01m")
	case diff < 1*Hour:
		buff.WriteString(fmt.Sprintf("%02dm", diff/Minute))
		diff -= diff / Minute * Minute
	case diff < 2*Hour:
		diff -= 1 * Hour
		buff.WriteString("01h")
	case diff < 1*Day:
		buff.WriteString(fmt.Sprintf("%02dh", diff/Hour))
		diff -= diff / Hour * Hour
	case diff < 2*Day:
		diff -= 1 * Day
		buff.WriteString("01d")
	case diff < 1*Week:
		buff.WriteString(fmt.Sprintf("%02dd", diff/Day))
		diff -= diff / Day * Day
	case diff < 2*Week:
		diff -= 1 * Week
		buff.WriteString("1w")
	case diff < 1*Month:
		buff.WriteString(fmt.Sprintf("%01dw", diff/Week))
		diff -= diff / Week * Week

	case diff < 2*Month:
		diff -= 1 * Month
		buff.WriteString("01M")
	case diff < 1*Year:
		buff.WriteString(fmt.Sprintf("%02dM", diff/Month))
		diff -= diff / Month * Month

	case diff < 2*Year:
		diff -= 1 * Year
		buff.WriteString("01y")
	default:
		buff.WriteString(fmt.Sprintf("%02dy", diff/Year))
		diff = 0
	}
	return diff
}

// TimeSincePro calculates the time interval and generate full user-friendly string.
func TimeSincePro(then time.Time) string {
	now := time.Now()
	diff := uint64(now.UnixNano() - then.UnixNano())

	if then.After(now) {
		return "future"
	}

	return TimeSinceDuration(diff)
}

// TimeSincePro calculates the time interval and generate full user-friendly string.
func TimeSinceDuration(then uint64) string {
	var (
		timeStr bytes.Buffer
		diff    = then
	)

	for {
		diff = computeTimeDiff(&timeStr, diff)
		if diff == 0 {
			break
		} else {
			timeStr.WriteRune(' ')
		}
	}
	if timeStr.Len() == 0 {
		return "instant!"

	} else {
		return timeStr.String()
	}
}

const (
	Byte  = 1
	KByte = Byte * 1024
	MByte = KByte * 1024
	GByte = MByte * 1024
	TByte = GByte * 1024
	PByte = TByte * 1024
	EByte = PByte * 1024
)

var bytesSizeTable = map[string]uint64{
	"b":  Byte,
	"kb": KByte,
	"mb": MByte,
	"gb": GByte,
	"tb": TByte,
	"pb": PByte,
	"eb": EByte,
}

func logn(n, b float64) float64 {
	return math.Log(n) / math.Log(b)
}

func humanateBytes(s uint64, base float64, sizes []string) string {
	if s < 10 {
		return fmt.Sprintf("%dB", s)
	}
	e := math.Floor(logn(float64(s), base))
	suffix := sizes[int(e)]
	val := float64(s) / math.Pow(base, math.Floor(e))
	f := "%.0f"
	if val < 10 {
		f = "%.1f"
	}

	return fmt.Sprintf(f+"%s", val, suffix)
}

// FileSize calculates the file size and generate user-friendly string.
func FileSize(s uint64) string {
	sizes := []string{"B", "KB", "MB", "GB", "TB", "PB", "EB"}
	return humanateBytes(s, 1024, sizes)
}

// Subtract deals with subtraction of all types of number.
func Subtract(left interface{}, right interface{}) interface{} {
	var rleft, rright int64
	var fleft, fright float64
	var isInt bool = true
	switch left.(type) {
	case int:
		rleft = int64(left.(int))
	case int8:
		rleft = int64(left.(int8))
	case int16:
		rleft = int64(left.(int16))
	case int32:
		rleft = int64(left.(int32))
	case int64:
		rleft = left.(int64)
	case float32:
		fleft = float64(left.(float32))
		isInt = false
	case float64:
		fleft = left.(float64)
		isInt = false
	}

	switch right.(type) {
	case int:
		rright = int64(right.(int))
	case int8:
		rright = int64(right.(int8))
	case int16:
		rright = int64(right.(int16))
	case int32:
		rright = int64(right.(int32))
	case int64:
		rright = right.(int64)
	case float32:
		fright = float64(left.(float32))
		isInt = false
	case float64:
		fleft = left.(float64)
		isInt = false
	}

	if isInt {
		return rleft - rright
	} else {
		return fleft + float64(rleft) - (fright + float64(rright))
	}
}

// EllipsisString returns a truncated short string,
// it appends '...' in the end of the length of string is too large.
func EllipsisString(str string, length int) string {
	if len(str) < length {
		return str
	}
	return str[:length-3] + "..."
}

func AspectRatio(srcRect image.Point, toResize uint64) image.Point {
	w, h := int(toResize), getRatioSize(int(toResize), srcRect.Y, srcRect.X)
	if srcRect.X < srcRect.Y {
		w, h = getRatioSize(int(toResize), srcRect.X, srcRect.Y), int(toResize)
	}
	return image.Point{w, h}
}
func getRatioSize(a, b, c int) int {
	d := a * b / c
	return (d + 1) & -1
}

func Basename(s string) string {
	n := strings.LastIndexByte(s, '.')
	if n >= 0 {
		return s[:n]
	}
	return s
}
